package light;
import java.awt.geom.Point2D;
import java.util.Random;


public class SimServer {

	//-----Server parameters-------
	/**
	 * Real-time length of one time step in ms
	 */
	public static final double STEPLENGTH = 100;
	
	/** Dimensions of the square field */
	public static final int FIELDDIM = 200;
	//public static final Point FIELDDIM = new Point(200,200);
	
	//------------Communication constants
	public static final int SETSPEED = 100;
	
	//-------------Simulator constants------------
	public static final double LIGHTSOURCE_PLACEMENT_RADIUS = 80;
	public static final double LIGHTSOURCE_SIZE = 10;
	
	/** Location of the light source */
	public Point2D.Double targetlocation;
	/** Reference to parent object */
	private Light parent;
	public boolean succeeded = false;
	
	public SimServer(Light parent, long seed) {
		this.parent = parent;
		
		initfield(seed);
	}
	
	private void initfield(long seed) { 	
		
		Random generator = new Random(seed);
		
		//place light source on a random point on a circle
		double angle = Math.toRadians(generator.nextInt(360));
		targetlocation = new Point2D.Double(LIGHTSOURCE_PLACEMENT_RADIUS * Math.cos (angle),LIGHTSOURCE_PLACEMENT_RADIUS * Math.sin (angle)    );
		//place target randomly outside this
		targetlocation = new Point2D.Double((LIGHTSOURCE_PLACEMENT_RADIUS+generator.nextInt(25)) * Math.cos (angle),(LIGHTSOURCE_PLACEMENT_RADIUS+generator.nextInt(25)) * Math.sin (angle)    );
	}
	
	public void calculateAll() {
		//calculate intentions
		calcint (parent.robot);
		//check if objective is completed
		if (getDistance(targetlocation, parent.robot.getPosition()) <= Robot.ROBOT_DIAMETER/2+1) {
			succeeded = true;
		}
	}
	
	private void calcint(Robot robot) {
		Intention intent =  consumeIntention(robot);
		switch (intent.intId) {
		case SimServer.SETSPEED : {//only one implemented so far
			double leftspeed = intent.param1;
			double rightspeed = intent.param2;
			//re-scale power to have negative as well (backward)
			leftspeed = (leftspeed*2*Robot.MOTOR_SPEED_MAX)-Robot.MOTOR_SPEED_MAX;
			rightspeed = (rightspeed*2*Robot.MOTOR_SPEED_MAX)-Robot.MOTOR_SPEED_MAX;
			
			//System.out.println (leftspeed+","+rightspeed);
			
			//calculate movement
			double time = STEPLENGTH/1000;
			double diff = leftspeed-rightspeed;
			double sum = leftspeed+rightspeed;
			double theta_speed = ( (double)(Robot.ROBOT_WHEEL_DIAMETER)/2.0)/((double)(Robot.ROBOT_DIAMETER))*diff*2;
			
			//angular change
			robot.setBodyDirection(robot.getBodyDirection()+Math.toDegrees(theta_speed*time));

			double x = Robot.ROBOT_WHEEL_DIAMETER/4.0*sum*Math.sin( Math.toRadians(robot.getBodyDirection()) )*time;
			
			double y = Robot.ROBOT_WHEEL_DIAMETER/4.0*sum*(-1)*Math.cos( Math.toRadians(robot.getBodyDirection()) )*time;
			
			//calculate new position
			Point2D.Double newpos = new Point2D.Double (robot.getPosition().x+x,robot.getPosition().y+y);
			
			if (ValidatePoint(newpos))	{ //object is not in blocked zone

				robot.setPosition(newpos); //everything is fine, move
				
			} else { //robot would leave zone
				//do nothing
			}
			
			break;
		}
		}
	}
	
	/**
	 * Calculates the length of an origo-based vector
	 * @return
	 */
	public static double getLength(Point2D.Double p1) {
		return Math.sqrt((Math.pow(p1.x, 2) + Math.pow(p1.y, 2)));
	}
	
	/** Returns false if the point is outside of the zone */
	private boolean ValidatePoint(Point2D.Double pos) {
		//check boundary
		if ((pos.x < - (FIELDDIM/2-Robot.ROBOT_DIAMETER/2)  ) || (pos.y<-(FIELDDIM/2-Robot.ROBOT_DIAMETER/2)) || (pos.x>(FIELDDIM/2-Robot.ROBOT_DIAMETER/2)) || (pos.y > (FIELDDIM/2-Robot.ROBOT_DIAMETER/2))) {
			return false;
		}
		else return true;
	}
	
	private Intention consumeIntention(Robot robot) {
		Intention intent;
		if (robot.buffer.size() == 0) { intent = null; System.err.println ("no intention");}
		else {
			intent = robot.buffer.remove(0);
		}
		return intent;
	}

	/**
	 * Calculates the difference between two angles given in degrees. 
	 * A positive result indicates that the second angle is further counterclockwise.  
	 * @param alpha first angle
	 * @param beta second angle
	 * @return difference in degrees between -180 and +180
	 */
	static public double angleDiffDegrees(double alpha,double beta) {
		double phi = (beta % 360)-(alpha % 360);
		
		if (phi > 180) phi-= 360;
		if (phi < -180) phi+= 360;
		
		return phi;
	}
	
	//static double wilsx=0.0;
	
	public void getVisuals(Robot robot) {
		double sensorleft = 0, sensorright = 0;	
		
		Point2D.Double pos = robot.getPosition();
		double dir = robot.getBodyDirection();
		dir = fixangle (dir);
		
		//left light sensor
		Point2D.Double sensorcenter = new Point2D.Double (pos.x+Math.sin(Math.toRadians(dir-90))*(Robot.ROBOT_DIAMETER/2.0), pos.y-Math.cos(Math.toRadians(dir-90))*(Robot.ROBOT_DIAMETER/2.0));
		double distance = getDistance (targetlocation,sensorcenter);
		double dx = targetlocation.x - sensorcenter.x;
		double dy = targetlocation.y - sensorcenter.y;
		
		double angle = Math.toDegrees(Math.atan2(dx,-dy));
		angle = fixangle (angle);
		
		if ((Math.abs(angleDiffDegrees(angle,dir))<Robot.LIGHT_SENSOR_ANGLE/2) && (distance <= Robot.LIGHT_SENSOR_RANGE)) {
		//if ((angle > (dir-Robot.LIGHT_SENSOR_ANGLE/2)) && (angle < (dir+Robot.LIGHT_SENSOR_ANGLE/2)) && (distance <= Robot.LIGHT_SENSOR_RANGE)) {
			//in range
			sensorleft = distance;
		}
		
		//right sensor
		sensorcenter = new Point2D.Double (pos.x+Math.sin(Math.toRadians(dir+90))*(Robot.ROBOT_DIAMETER/2.0), pos.y-Math.cos(Math.toRadians(dir+90))*(Robot.ROBOT_DIAMETER/2.0));
		distance = getDistance (targetlocation,sensorcenter);
		dx = targetlocation.x - sensorcenter.x;
		dy = targetlocation.y - sensorcenter.y;
		//angle = Math.toDegrees(Math.asin(dx/distance));
		angle = Math.toDegrees(Math.atan2(dx,-dy));
		angle = fixangle (angle);
		
		//if ((angle > (dir - Robot.LIGHT_SENSOR_ANGLE/2)) && (angle < (dir+Robot.LIGHT_SENSOR_ANGLE/2)) && (distance <= Robot.LIGHT_SENSOR_RANGE)) {
		if ((Math.abs(angleDiffDegrees(angle,dir))<Robot.LIGHT_SENSOR_ANGLE/2) && (distance <= Robot.LIGHT_SENSOR_RANGE)) {
			//in range
			sensorright = distance;
	//		System.out.println ("I CAN SEE IT dist: "+distance+" angle: "+angle+"tloc.x: "+targetlocation.x+" sc: "+sensorcenter.x);
		}
		else {
	//		System.out.println ("dist: "+distance+" angle: "+angle+" tloc.x: "+targetlocation.x+" sc: "+sensorcenter.x);
	//		System.out.println ("dx: "+dx+" dy: "+dy+" dir: "+dir);
	//		System.out.println();
			
		}
		robot.getController().setSensorValues(new double[]{sensorleft,sensorright});
		
	}
	
	private double fixangle (double angle) {
		if (angle >= 0) {
			return angle % 360;
		}
		else {
			angle = angle % 360;
			return 360+angle;
		}
	}
	
	public static double getDistance(Point2D.Double obj1, Point2D.Double obj2) {
		return Math.sqrt((Math.pow((obj1.x-obj2.x), 2)) + (Math.pow((obj1.y-obj2.y), 2)));
	}
}
