package nnracer;
/*
 * Copyright (C) 2009 Istvan Fehervari, Wilfried Elmenreich
 * Original project page: http://www.frevotool.tk
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License Version 3 as published
 * by the Free Software Foundation http://www.gnu.org/licenses/gpl-3.0.txt
 *
 * There is no warranty for this free software. The GPL requires that 
 * modified versions be marked as changed, so that their problems will
 * not be attributed erroneously to authors of previous versions.
 */
import interfaces.IProblem;
import interfaces.IRepresentation;
import interfaces.PropEntry;

import java.util.Hashtable;
import java.util.Random;
import java.util.Vector;


public class NNRacer extends IProblem {
	
	public int[] obstaclePattern;
	private Random generator = new Random();
	private float carProbability;
	public int [] baseline;
	private int lanenum, lookahead;
	private int evalTimes, maxovertake;
	
	boolean printGameSteps = false;

	@Override
	public double getResult(IRepresentation[] candidates) {
		if (candidates.length != 1) throw new Error ("Should pass only one representation!");

		IRepresentation driver = candidates[0];
				
		//extract properties
	//	long randomseed = Long.parseLong(properties.get("randomSeed").getValue());
		lanenum = Integer.parseInt(getProperties().get("laneNumber").getValue());
		lookahead = Integer.parseInt(getProperties().get("lookAhead").getValue());
		carProbability = Float.parseFloat(getProperties().get("carProbability").getValue());
		maxovertake = Integer.parseInt(getProperties().get("maxOvertakeNumber").getValue());
		evalTimes = Integer.parseInt(getProperties().get("evaluationTimes").getValue());		
		
		int sum = 0;
		for (int r=0;r<evalTimes;r++) {
			generator = new Random(0x5EED + r);
			sum = sum + getEval(driver);
		}
		int res = (sum/evalTimes);
		return (res);
	}
	
	private int getEval(IRepresentation driver) {
		boolean isColloided = false;
		int finishedCycles = 0;
		int position = 0; //vertical position of the car; can be -1, 0, 1
		
		baseline = new int[lanenum];
		obstaclePattern = new int [lanenum*lookahead];
		
		//initialize obstacle pattern
		generateInput (true);
		
		while (!isColloided) {
			if (finishedCycles > maxovertake) break;
			//increase difficulty
			setDifficulity(finishedCycles);
			//feed input to the network
			Vector<Float> input = new Vector<Float>();
			
			if (printGameSteps) System.out.print("in: ");
			for (int in = 0; in < lanenum; in++) {
				input.add( (float)obstaclePattern[in] );
				if (printGameSteps) System.out.print(obstaclePattern[in]+",");
			}
			input.add ( (float) position); //add position
			if (printGameSteps) System.out.print(position);
			
			//get output
			Vector<Float> output = new Vector<Float>();
			output.setSize(1);
			
			output = driver.getOutput(input);
			
			//get new position from output
			if (printGameSteps) System.out.print (" output: "+output.get(0)+", ");
			
			position = getPosition (output.get(0));
			
			if (printGameSteps) System.out.println (position);
			
			//check collision TELEPORTATION MODEL
			if (obstaclePattern[position] == 1) isColloided = true; //collision happened
			else { //no collision, propagate cars								
				generateInput(false);
				finishedCycles++;
			}			
		}
		return finishedCycles;
	}
	
	private void setDifficulity(int cyc) {
		if ((cyc > 50) && (cyc <100)   ) carProbability = 0.6f;
		else if ((cyc >= 100) && (cyc <150)   ) carProbability = 0.7f;
		else if ((cyc >= 150) && (cyc <200)) carProbability = 0.8f;
		else if (cyc >= 200) carProbability = 0.9f;
	}
	
	/** Converts the output of the representation to position */
	private int getPosition(float out) {
		// out is always in the range of -1 .. +1
		float l = (2.0f)/(float)lanenum;
		float k = java.lang.Math.abs(-1-out);
		int pos = (int)((k/l));
		if (pos > lanenum-1) pos = lanenum-1;
		
		return pos;
	}
	
	/**
	 * Generates binary input grid. Depends on the car probability 
	 * @param init
	 */
	public void generateInput(boolean init) {
		int l = lanenum;
		if (init) {//generate new input grid
			for (int row = 0; row < lookahead; row++) { //generate each row
				int emptyGrid = generator.nextInt(l); //randomize one empty spot to ensure that the task is solvable
				for (int i=0;i<l;i++) {
					if ( i != emptyGrid  ) { 
						if (generator.nextDouble() < carProbability ) obstaclePattern[i+(l*row)] = 1;
						else obstaclePattern[i+(l*row)] = 0;
					}
					else obstaclePattern[i+(l*row)] = 0;
				}
			}			
			
		}
		else { //progress line one to line two, generate new line one
			if (lookahead == 1) { //if there is only one visible line we have to generate it instead of propagating
				int emptyGrid = generator.nextInt(l); //randomize one empty spot to ensure that the task is solvable
				for (int i=0;i<l;i++) {
					if ( i != emptyGrid  ) { 
						if (generator.nextDouble() < carProbability ) obstaclePattern[i] = 1;
						else obstaclePattern[i] = 0;
					}
					else obstaclePattern[i] = 0;
				}
			}
			else { //we have more lines
				//propagate information row-1 times
				for (int r = 0; r<lookahead-1; r++) {
					for (int k = 0;k<lanenum;k++) {
						obstaclePattern[k+(l*r)] = obstaclePattern[k+(l*(r+1))];
					}
				}								
				//create new "upper" line
				int emptyGrid = generator.nextInt(l); //randomize one empty spot to ensure that the task is solvable
				for (int i=0;i<l;i++) {
					if ( i != emptyGrid  ) { 
						if (generator.nextDouble() < carProbability ) obstaclePattern[i+(l*(lookahead-1) )] = 1;
						else obstaclePattern[i+(l*(lookahead-1))] = 0;
					}
					else obstaclePattern[i+(l*(lookahead-1))] = 0;
				}
			}
			
		}
	}

	@Override
	public void replayWithVisualization(IRepresentation[] candidates) {
		System.out.println ("Result: "+getResult(candidates));
		
	}

	@Override
	public Hashtable<String, PropEntry> adjustRequirements(Hashtable<String, PropEntry> requirements,
			Hashtable<String, PropEntry> properties) {
		PropEntry ln = properties.get("laneNumber");
		int laneNumber = Integer.parseInt(ln.getValue());
		PropEntry la = properties.get("lookAhead");
		int lookAhead = Integer.parseInt(la.getValue());
		
	/*	INPUT_NODES = (lanenum*lookahead)+1;
		OUTPUT_NODES = 1;*/
		
		PropEntry inputn = requirements.get("inputnumber");
		inputn.setValue(Integer.toString((laneNumber*lookAhead)+1));
		
		return requirements;
	}

}
