package fullyMeshedNet;
/*
 * Copyright (C) 2011 Istvan Fehervari, Wilfried Elmenreich
 * Original project page: http://www.frevotool.tk
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License Version 3 as published
 * by the Free Software Foundation http://www.gnu.org/licenses/gpl-3.0.txt
 *
 * There is no warranty for this free software. The GPL requires that 
 * modified versions be marked as changed, so that their problems will
 * not be attributed erroneously to authors of previous versions.
 */

import interfaces.IRepresentation;
import interfaces.PropEntry;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Vector;

import main.FrevoMain;

import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.Node;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.XMLWriter;

import utils.NESRandom;


public class FullyMeshedNet extends IRepresentation {
	
	protected NESRandom generator;
	protected NESRandom randomInput = new NESRandom(1234);
	protected float[][] weight;
	protected float[] bias;
	protected float[] randombias;
	protected float[] output;
	protected float[] activation;
	protected int rank;
	protected int nodes;
	protected int input_nodes;
	protected int output_nodes;
	protected int hidden_nodes;
	protected float weight_range;
	protected float bias_range;
	protected float random_bias_range;
	protected int iterations;
	protected boolean random_source; //if true, each neuron has a second bias which is subject to random input
	protected boolean variable_mutation_rate; //if true, the mutation rate itself can be mutated and inherited
	protected float mutationRate; //this variable is only effective when variable_mutation_rate is active
	final float initialMutationRate=1.0f;
	
	@Override
	public void generate(int inputnumber, int outputnumber, NESRandom r) {
		generator = r;
		input_nodes = inputnumber;
		output_nodes = outputnumber;
		// load properties from properties
		loadProperties();
		//generate structure
		this.weight = new float[this.nodes][this.nodes];
		this.bias = new float[this.nodes];
		this.randombias = new float[this.nodes];
		this.output = new float[this.nodes];
		this.activation = new float[this.nodes];
		mutationRate = initialMutationRate;
		if (r != null) randomizeWB(); //this step is not needed for cloning, otherwise random is not null
	}
	
	private void loadProperties() {		
		PropEntry pIterations = getProperties().get("iterations");
		iterations = Integer.parseInt(pIterations.getValue());
		
		PropEntry wr = getProperties().get("weight_range");
		this.weight_range = Float.parseFloat(wr.getValue());
		PropEntry br = getProperties().get("bias_range");
		this.bias_range = Float.parseFloat(br.getValue());
		PropEntry hn = getProperties().get("hiddenNodes");
		hidden_nodes = Integer.parseInt(hn.getValue());
		PropEntry rs = getProperties().get("random_source");
		random_source = Boolean.parseBoolean(rs.getValue());
		PropEntry rbr = getProperties().get("random_bias_range");
		this.random_bias_range = Float.parseFloat(rbr.getValue());
		PropEntry vmr = getProperties().get("variable_mutation_rate");
		variable_mutation_rate = Boolean.parseBoolean(vmr.getValue());
		
		nodes = input_nodes+output_nodes+hidden_nodes;
	}
	
	private void randomizeWB()
	{		
		for(int i = input_nodes;i<nodes;i++)
		{
			for(int j = 0;j<nodes;j++)
			{
				this.weight[j][i] = rand_range(weight_range);
			}
			this.output[i] = 0;	//*f
			this.bias[i] = rand_range(bias_range);
			
			if (random_source)
				this.randombias[i] = (float)(1 - randomInput.nextDouble() * 2) * random_bias_range;
		}
	}
	
	//*f
	private float rand_range(float border)
	{
		double val;
		val = generator.nextDouble() * 2 * border - border;  //*f

		return (float)val;	
	}

	@Override
	public int getNumberofMutationMethods() {
		return 2;
	}
	
	@Override
	public int getNumberofXoverMethods() {
		return 1;
	}

	@Override
	public IRepresentation mutate(float severity, int probability, int method) {
		float usedMutationrate;
		
		if (variable_mutation_rate) {
			//use local muationrate
			usedMutationrate = mutationRate;
		}
		else {
			//use global mutationrate (severity)
			usedMutationrate = severity;
		}
		
		this.mutateBias(usedMutationrate, probability, method);
		this.mutateWeight(usedMutationrate, probability, method);
		if (random_source) 
			this.mutateRandomBias(100,(int)(usedMutationrate*100), method);
				
		if (variable_mutation_rate) {
			mutateMutationrate();
		}
		return this;
	}
	
	private void mutateBias(float severity, int p, int method)
	{		
		float rate = severity/100*bias_range; //use bias_range as 100%
		float probability = p;
		
		for(int i = input_nodes;i<bias.length;i++)
		{
			if (generator.nextInt(100) < probability) {
				switch(method) {
				case 1:
					this.bias[i] += rand_range(rate);
					break; 
				case 2: // smart method, facilitates the mutation to fine-tune small parameters
					this.bias[i] += rand_range((int)Math.max(this.bias[i]/5.0,10)); // 20% mutation
					break;
				default: 
					this.bias[i] += rand_range(rate);
					break; 
				}
			}
		}
	}
	
	private void mutateRandomBias(float severity, int p, int method)
	{		
		float rate = severity/100* random_bias_range;
		float probability = p;
		
		for(int i = input_nodes;i<bias.length;i++)
		{
			if (generator.nextInt(100) < probability) {
				switch(method) {
				case 1:
					this.randombias[i] += rand_range(rate);
					break; 
				case 2: // smart method, facilitates the mutation to fine-tune small parameters
					this.randombias[i] += rand_range((int)Math.max(this.randombias[i]/5.0,10)); // 20% mutation
					break;
				default: 
					this.randombias[i] += rand_range(rate);
					break; 
				}
			}
		}
	}
	
	private void mutateWeight(float severity, int p, int method)
	{
		
		float rate = severity/100*weight_range; //use weight_range as 100%
		float probability = p;
		
		//System.out.println ("mutating weight rate:"+rate+" prob:"+p);
		
		for(int i = 0;i<bias.length;i++)
		{
			for(int j = input_nodes;j<weight.length;j++)
			{
				if (generator.nextInt(100) < probability)
					switch(method) {
					case 1:
						this.weight[i][j] += rand_range(rate);
						break;
					case 2: // smart method, facilitates the mutation to fine-tune small parameters
					//	float change=rand_range((int)Math.max(this.weight[i][j]/5.0,10)); // 20% mutation
						//System.out.print("ch("+i+","+j+"):"+change+" ");
						this.weight[i][j] += rand_range((float) Math.max(this.weight[i][j]/5.0,0.1)); // 20% mutation or at least range of 0.1				
						break;
					default:
						this.weight[i][j] += rand_range(rate);
						break;
					}
			}
		}
	}

	private void mutateMutationrate()
	{
		mutationRate += rand_range((float) Math.max(mutationRate/5.0,0.1)); // 20% mutation or at least range of 0.1	
	}
	
	public String getUniqueName() {
		double sum = 0;
		//add bias
		for (int i=0;i<bias.length;i++) {
			sum += bias[i];
		}
		//add weights
		for(int i = input_nodes;i<nodes;i++)
		{
			for(int j = 0;j<nodes;j++)
			{
				sum += this.weight[j][i];
			}
		}
		String res = Double.toString(sum);
		int resint = res.hashCode();
		return Integer.toHexString(resint & 0xFFFFF);
	}
	
	private int rndIndex(int range)
	{
		if(range < 1)
			return 0;
		else
			return generator.nextInt(Integer.MAX_VALUE) % range;
	}

	@Override
	public IRepresentation xOverWith(IRepresentation other, int method) {
		if(!(other instanceof FullyMeshedNet))
			throw new IllegalArgumentException("Xover between different network classes not possible!");
		else {
			// modify this representation
			if (method == 1) {
				int noninodes = nodes-input_nodes;
				int count = rndIndex(noninodes-1)+1; //minimum 1, maximum all but 1
				int start = input_nodes + rndIndex(noninodes);
				
				FullyMeshedNet offspring = (FullyMeshedNet) this;
				FullyMeshedNet father = (FullyMeshedNet) other;
				
				for(int i = start;i<start+count;i++)
				{
					if(i>=nodes)
						break;
					
					for(int j = 0;j<nodes;j++)
						offspring.weight[j][i] = father.weight[j][i];
					offspring.bias[i] = father.bias[i];
					if (random_source) offspring.randombias[i] = father.randombias[i]; 
				}
				//return offspring;
			}
		/*	else if (method == 2) {
				not yet implemented
			}*/
			else this.xOverWith(other, method);
		}
		
		return this;
	}
	
	public float activate (float x)
	{		
		if(x>=1)
			return 1;
		else if(x<=0)
			return 0;
		else 
			return x;
		
	}
	
	public void reset()
	{
		for(int i=0;i<nodes;i++)
			this.output[i] = 0;
	}
	
	public Vector<Float> getOutput(Vector<Float> input) {
		for (int i=0;i<iterations-1;i++) {
			getStep(input);
		}
		return getStep (input);
	}

	/** Provides the networks output for the given input*/
	public Vector<Float> getStep(Vector<Float> input) {
		if(input.size() != this.input_nodes)
			throw new IllegalArgumentException("Input vector size inappropriate!\ninput size: "+input.size()+" while input nodes are "+this.input_nodes);
		
		for(int i = 0;i<input.size();i++)
			this.output[i] = input.get(i);
	
		for(int i = input_nodes;i<nodes;i++)
		{
			float sum = 0;
			for(int j = 0;j<nodes;j++)
			{
				sum += this.weight[j][i] * this.output[j];
				//System.out.println("sum="+sum);
			}
			this.activation[i] = this.bias[i] + sum;
			if (random_source) 
				this.activation[i]+= rand_range(randombias[i]);
		}
		
		for(int i = input_nodes;i<nodes;i++)
		{
			this.output[i] = activate(this.activation[i]);
		}

		Vector<Float> outputVector = new Vector<Float>();
		
		for(int i = nodes-output_nodes;i<nodes;i++)
		{
			outputVector.add(this.output[i]);
		}
		return outputVector;
	}

	@Override
	public void setRank(int rank) {
		this.rank=rank;	
	}

	@Override
	public double diffTo(IRepresentation o) {
		double diff=0;
		
		if(!(o instanceof FullyMeshedNet))
			throw new IllegalArgumentException("diffTo between different network classes not possible!");
		else
		{
			FullyMeshedNet network = (FullyMeshedNet) o;
			
			for(int i = 0;i<this.weight.length;i++ )
			{
				for(int j = 0;j<this.weight[i].length;j++)
				{
					diff += Math.abs(network.weight[i][j] - this.weight[i][j]); 
				}
			}
			for(int i =0;i<this.bias.length;i++)
				diff += Math.abs(network.bias[i] - this.bias[i]);
			if (random_source)
				for(int i =0;i<this.randombias.length;i++)
					diff += Math.abs(network.randombias[i] - this.randombias[i]);
		}
		
		return diff;
	}

	public IRepresentation clone() {
		FullyMeshedNet c = new FullyMeshedNet();
		c.addProperties(getProperties());
		c.generate(input_nodes, output_nodes, new NESRandom(generator.nextLong()));
		
		for(int i = 0;i<this.weight.length;i++ )
		{
			for(int j = 0;j<this.weight[0].length;j++)
			{
				c.weight[i][j] = this.weight[i][j]; 
			}
		}
		
		c.bias = this.bias.clone();
		c.output = this.output.clone();
		c.activation = this.activation.clone();
		c.rank = this.rank;
		c.nodes = this.nodes;
		c.input_nodes = this.input_nodes;
		c.output_nodes = this.output_nodes;
		c.hidden_nodes = this.hidden_nodes;
		c.weight_range = this.weight_range;
		c.bias_range = this.bias_range;
		c.iterations = this.iterations;
		c.random_source = this.random_source;
		if (random_source) {
			c.randombias = this.randombias.clone();
			c.random_bias_range = this.random_bias_range;
		}
		
		c.setFitness(this.getFitness());
		
		//res.addProperties(getProperties());
				
		String n1= this.getUniqueName();
		String n2= c.getUniqueName();
		
		if (!n1.equals(n2)) System.out.println(this.getUniqueName() + "cloned to" + c.getUniqueName());
		
		return c;
	}
	
//	public IRepresentation clone3() { 
//		FullyMeshedNet res = new FullyMeshedNet();
//		res.addProperties(getProperties());
//		res.generate(input_nodes, output_nodes, new NESRandom(generator.nextLong()));
//		for(int i = 0;i<this.weight.length;i++ )
//		{
//			for(int j = 0;j<this.weight[0].length;j++)
//			{
//				res.weight[i][j] = this.weight[i][j]; 
//			}
//		}
//		
//		res.bias = this.bias.clone();
//		res.output = this.output.clone();
//		res.activation = this.activation.clone();
//		res.setFitness(this.getFitness());
//	/* 
//		res.iteration = this.iteration;
//		res.gahist = this.gahist.clone();*/
//		res.rank = this.rank;
//		
//		String n1= this.getUniqueName();
//		String n2= res.getUniqueName();
//		
//		if (!n1.equals(n2)) System.out.println(this.getUniqueName() + "cloned to" + res.getUniqueName());
//		
//		return res;
//	}		
	
	public float getWeight(int nr, int from)
	{
		if(nr >= nodes)
			throw new IllegalArgumentException("Id "+nr+" is not an node!");
		{
			if(from >= nodes)
				throw new IllegalArgumentException("Id "+from+" is not an node!");
			else
				return this.weight[from][nr];
		}
	}
	
	/** Method used for saving the properties of this component*/
	public Element exportXml (Element element) {
		Element nn = element.addElement("FullyMeshedNet");
		nn.addAttribute("input_nodes",String.valueOf(this.input_nodes));
		nn.addAttribute("output_nodes",String.valueOf(this.output_nodes));
		nn.addAttribute("nodes",String.valueOf(this.nodes));
		nn.addAttribute("weight_range",String.valueOf(this.weight_range));
		nn.addAttribute("bias_range",String.valueOf(this.bias_range));
		nn.addAttribute("random_bias_range",String.valueOf(this.random_bias_range));
		nn.addAttribute("random_source",String.valueOf(this.random_source));
		nn.addAttribute("variable_mutation_rate",String.valueOf(this.variable_mutation_rate));
		
		if (variable_mutation_rate)
			nn.addAttribute("mutation_rate",String.valueOf(this.mutationRate));
		
	/*	Element simulation = nn.addElement("simulation");
		simulation.addAttribute("iteration",String.valueOf(this.iteration));
		simulation.addAttribute("score",String.valueOf(this.score));*/
		
		Element dnodes = nn.addElement("nodes");
		
		for(int i = 0;i<this.nodes;i++)
		{
			Element node;
			if(i<input_nodes)
			{
				node = dnodes.addElement("node");
				node.addAttribute("nr", String.valueOf(i));
				node.addAttribute("type", "inputNode");
			}
			else
			{
				node = dnodes.addElement("node");
				node.addAttribute("nr", String.valueOf(i));
				node.addAttribute("type", "node");
				Element bias = node.addElement("bias");
				bias.addText(String.valueOf(this.bias[i]));
				
				if (random_source) {
					Element randombias = node.addElement("randombias");
					randombias.addText(String.valueOf(this.randombias[i]));					
				}
					
				
				Element weights = node.addElement("weights");
				for(int j = 0;j<this.nodes;j++)
				{
					Element weight = weights.addElement("weight");
					weight.addAttribute("from",String.valueOf(j));
					weight.addText(String.valueOf(this.getWeight(i, j)));
				}
			}
		}
				
		return nn;
	}
	
	public IRepresentation loadFromXML(Node nd) {
		try {
			if (this.generator == null) this.generator = new NESRandom();
			//Add properties
			loadProperties();
			this.input_nodes = Integer.parseInt(nd.valueOf("./@input_nodes"));
			this.output_nodes = Integer.parseInt(nd.valueOf("./@output_nodes"));
			this.nodes = Integer.parseInt(nd.valueOf("./@nodes"));
			this.weight_range = Float.parseFloat(nd.valueOf("./@weight_range"));
			this.bias_range = Float.parseFloat(nd.valueOf("./@bias_range"));
			this.random_bias_range = Float.parseFloat(nd.valueOf("./@random_bias_range"));
			this.random_source = Boolean.parseBoolean(nd.valueOf("./@random_source"));
			
			this.variable_mutation_rate = Boolean.parseBoolean(nd.valueOf("./@variable_mutation_rate"));
			if (variable_mutation_rate)
				this.mutationRate = Float.parseFloat(nd.valueOf("./@mutation_rate"));
			
			//this.iteration = Integer.parseInt(nd.valueOf("./simulation/@iteration"));
			//this.score = Integer.parseInt(nd.valueOf("./simulation/@score"));
			this.activation = new float[this.nodes];
			this.output = new float[this.nodes];
			this.bias = new float[this.nodes];
			this.randombias = new float[this.nodes];
			this.weight = new float[this.nodes][this.nodes];
			
			
			Node dnodes = nd.selectSingleNode("./nodes");
			for(int nr = this.input_nodes;nr<this.nodes;nr++)
			{
				Node curnode = dnodes.selectSingleNode("./node[@nr='"+nr+"']");
				if(curnode == null)
					throw new IllegalArgumentException("CompleteNetwork: node tags inconsistent!" +
							"\ncheck 'nr' attributes and nodes count in nnetwork!");
				this.bias[nr] = Float.parseFloat(curnode.valueOf("./bias"));
				if (random_source) {
					this.randombias[nr] = Float.parseFloat(curnode.valueOf("./randombias"));			
				}
				
				Node dweights = curnode.selectSingleNode("./weights");
				for(int from=0;from<this.nodes;from++)
				{
					String ws = dweights.valueOf("./weight[@from='"+from+"']");
					if(ws.length() == 0)
						throw new IllegalArgumentException("CompleteNetwork: weight tags inconsistent!" +
								"\ncheck 'from' attributes and nodes count in nnetwork!");
					float val = Float.parseFloat(ws);
					this.weight[from][nr] = val;
				}
			}
			//this.gahist = new GAHistory(this.config, nd);
		}
		catch (NumberFormatException e)
		{
			throw new IllegalArgumentException("CompleteNetwork: NumberFormatException! Check XML File");
		}
		return this;
	}

	@Override
	public IRepresentation loadFromXML(Node nd, NESRandom r) {
		this.generator = r;
		return loadFromXML(nd);
	}

	@Override
	public Hashtable<String, String> getDetails() {
		Hashtable<String, String> result = new Hashtable<String, String>();
		result.put("input neurons", Integer.toString(this.input_nodes));
		result.put("output neurons", Integer.toString(this.output_nodes));
		result.put("hidden neurons", Integer.toString(nodes-input_nodes-output_nodes));
		result.put("weight range", Float.toString(this.weight_range));
		result.put("bias range", Float.toString(this.bias_range));
		if (random_source)
			result.put("random_bias range", Float.toString(this.random_bias_range));
		for (int n = 0; n < nodes;n++) {
			result.put("bias node "+n, Float.toString(this.bias[n]));
			if (random_source)
				result.put("random bias node "+n, Float.toString(this.randombias[n]));
			for(int from=0;from<this.nodes;from++) {
				result.put("weight from "+from+" to "+n, Float.toString(this.weight[from][n]));
			}
		}
		
		return result;
	}
	
	/**
	 * @return the bias
	 */
	public float getBias(int i) {
		return bias[i];
	}
	
	public int getNumberofNodes() {
		return nodes;
	}


	@Override
	public void exportToFile(File saveFile) {
		String extension = FrevoMain.getExtension(saveFile); 
		if (extension.equals("net") ) {
			System.out.println ("Exporting Pajek network file to "+saveFile.getName());
			try{
			    //Create file 
			    FileWriter fstream = new FileWriter(saveFile);
			    BufferedWriter out = new BufferedWriter(fstream);
			    out.write("*Vertices "+this.nodes);
			    out.newLine();
			    //input neurons
			    for (int i=1;i<input_nodes+1;i++) {
			    	out.write(i+" \"I"+i+"\"");
			    	out.newLine();
			    }
			    for (int h=input_nodes+1;h<nodes-output_nodes+1;h++) {
			    	out.write(h+" \"H"+(h-input_nodes)+"\"");
			    	out.newLine();
			    }
			    int a=1;
			    for (int o=nodes-output_nodes+1;o<nodes+1;o++) {
			    	out.write(o+" \"O"+a+"\"");
			    	out.newLine();
			    	a++;
			    }
			    
			    //Edges
			    out.write("*Edges");
			    out.newLine();
			    for (int n = 0; n< input_nodes-1;n++) {
			    	for (int p =n+1;p<input_nodes;p++) {
			    		if (n!=p) {
			    			out.write((n+1)+" "+(p+1)+" "+0);
			    			out.newLine();
			    		}
			    	}
			    }
			    
			    
			    for (int n = input_nodes; n < nodes;n++) {
			    	for (int from = 0;from <nodes;from++) {
			    		out.write((n+1)+" "+(from+1)+" "+weight[from][n]);
			    		out.newLine();
			    	}
				}
			    
			    //Close the output stream
			    out.close();
			    }catch (Exception e){
			      System.err.println("Error: " + e.getMessage());
			}
		}
		else if (extension.equals("xml")){
			System.out.println ("Saving to XML");
			
			Document doc = DocumentHelper.createDocument();
	        doc.addDocType("CompleteNetwork", null, System.getProperty("user.dir")+"//Components//Representations//CompleteNetwork//src//CompleteNetwork.dtd");
	        Element cnetwork = doc.addElement("CompleteNetwork");
	        this.exportXml(cnetwork);

            OutputFormat format = OutputFormat.createPrettyPrint();
            format.setLineSeparator(System.getProperty("line.separator"));

            try {
				saveFile.createNewFile();
				FileWriter out = new FileWriter(saveFile);
				BufferedWriter bw = new BufferedWriter(out);
				XMLWriter wr = new XMLWriter(bw, format);
				wr.write(doc);
				wr.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
	}

}
