/*
 * Copyright (C) 2009 Istvan Fehervari, Wilfried Elmenreich
 * Original project page: http://www.frevotool.tk
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License Version 3 as published
 * by the Free Software Foundation http://www.gnu.org/licenses/gpl-3.0.txt
 *
 * There is no warranty for this free software. The GPL requires that 
 * modified versions be marked as changed, so that their problems will
 * not be attributed erroneously to authors of previous versions.
 */
package threeLayerNetwork;
import interfaces.FileType;
import interfaces.IRepresentation;
import interfaces.PropEntry;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Vector;

import org.dom4j.Element;
import org.dom4j.Node;

import utils.NESRandom;

/**
 * A not recursive layered neural network
 * @author Wil
 */
public class ThreeLayerNetwork extends IRepresentation {

	protected NESRandom generator;
	protected float[][] weight;
	protected float[] bias;
	protected float[] output;
	protected float[] activation;
	protected int rank;
	protected int nodes;
	protected int input_nodes;
	protected int output_nodes;
	protected int hidden_nodes;
	protected float weight_range;
	protected float bias_range;
	protected int stepnumber;
	
	protected ArrayList<FileType> exportList = new ArrayList<FileType>();

	@Override
	public void generate(int inputnumber, int outputnumber, NESRandom r) {
		generator = r;
		input_nodes = inputnumber;
		output_nodes = outputnumber;
		// load properties	
		PropEntry snumber = getProperties().get("stepNumber");
		stepnumber = Integer.parseInt(snumber.getValue());
		
		PropEntry wr = getProperties().get("weight_range");
		this.weight_range = Float.parseFloat(wr.getValue());
		PropEntry br = getProperties().get("bias_range");
		this.bias_range = Float.parseFloat(br.getValue());
		PropEntry hn = getProperties().get("hiddenNodes");
		hidden_nodes = Integer.parseInt(hn.getValue());
		
		nodes = input_nodes+output_nodes+hidden_nodes;
		
		//generate structure
		this.weight = new float[this.nodes][this.nodes];
		this.bias = new float[this.nodes];
		this.output = new float[this.nodes];
		this.activation = new float[this.nodes];
		if (r != null) randomizeWB(); //this step is not needed for cloning, otherwise random is not null		
	}
	
	public String getUniqueName() {
		double sum = 0;
		//add bias
		for (int i=0;i<bias.length;i++) {
			sum += bias[i];
		}
		//add weights
		for(int i = input_nodes;i<nodes;i++)
		{
			for(int j = 0;j<nodes;j++)
			{
				sum += this.weight[j][i];
			}
		}
		String res = Double.toString(sum);
		int resint = res.hashCode();
		return Integer.toHexString(resint & 0xFFFFF);
	}
	
	private void randomizeWB()
	{		
		//set all weights and biases to 0
		for(int i = 0;i<nodes;i++)
		{
			for(int j = 0;j<nodes;j++)
			{
				this.weight[i][j] = 0;
			}
			this.output[i] = 0;
			this.bias[i] = 0;
		}			
			
		//randomize connections between hidden nodes and input nodes
		for(int i = input_nodes ; i<nodes - output_nodes ; i++) //all hidden nodes
		{
			for(int j = 0 ; j < input_nodes ; j++) //all input nodes
			{
				this.weight[j][i] = rand_range(weight_range);
			}
			this.output[i] = 0;
			this.bias[i] = rand_range(bias_range);
		}
		
		//randomize connections between hidden nodes and output nodes
		
		for(int i = nodes-output_nodes;i<nodes;i++) //all output nodes
		{
			for(int j = input_nodes ; j<nodes - output_nodes ; j++) //all hidden nodes
			{
				this.weight[j][i] = rand_range(weight_range);
			}
			this.output[i] = 0;
			this.bias[i] = rand_range(bias_range);
		}
		
		//no bias for input nodes
	}
	
	//*f
	private float rand_range(float border)
	{
		double val;
		val = generator.nextDouble() * 2 * border - border;  //*f

		return (float)val;	
	}

	@Override
	public int getNumberofMutationMethods() {
		return 2;
	}
	
	@Override
	public int getNumberofXoverMethods() {
		return 1;
	}

	@Override
	public IRepresentation mutate(float severity, int probability, int method) {
		this.mutateBias(severity, probability, method);
		this.mutateWeight(severity, probability, method);
		return this;
	}
	
	public void mutateBias(float severity, int p, int method)
	{		
		float rate = severity/100*bias_range; //use bias_range as 100%
		float probability = p;
		
		for(int i = input_nodes;i<bias.length;i++)
		{
			if (generator.nextInt(100) < probability) {
				switch(method) {
				case 1:
					this.bias[i] += rand_range(rate);
					break; 
				case 2: // smart method, facilitates the mutation to fine-tune small parameters
					this.bias[i] += rand_range((int)Math.max(this.bias[i]/5.0,10)); // 20% mutation
					break;
				default: 
					this.bias[i] += rand_range(rate);
					break; 
				}
			}
				
		}
	}
	
	public void mutateWeight(float severity, int p, int method)
	{
		float rate = severity/100*weight_range; //use weight_range as 100%
		float probability = p;
		
		for(int i = 0;i<bias.length;i++)
		{
			for(int j = input_nodes;j<weight.length;j++)
			{
				if (generator.nextInt(100) < probability)
					switch(method) {
					case 1:
						this.weight[i][j] += rand_range(rate);
						break;
					case 2: // smart method, facilitates the mutation to fine-tune small parameters
						this.weight[i][j] += rand_range((int)Math.max(this.weight[i][j]/5.0,10)); // 20% mutation
						break;
					default:
						this.weight[i][j] += rand_range(rate);
						break;
					}
					
			}
		}
	}
	
	private int rndIndex(int range)
	{
		if(range < 1)
			return 0;
		else
			return generator.nextInt(Integer.MAX_VALUE) % range;
	}

	@Override
	public IRepresentation xOverWith(IRepresentation other, int method) {
		if(!(other instanceof ThreeLayerNetwork))
			throw new IllegalArgumentException("Xover between different network classes not possible!");
		else {
			// modify this representation
			if (method == 1) {
				int noninodes = nodes-input_nodes;
				int count = rndIndex(noninodes-1)+1; //minimum 1, maximum all but 1
				int start = input_nodes + rndIndex(noninodes);
				
				ThreeLayerNetwork offspring = (ThreeLayerNetwork) this;
				ThreeLayerNetwork father = (ThreeLayerNetwork) other;
				
				for(int i = start;i<start+count;i++)
				{
					if(i>=nodes)
						break;
					
					for(int j = 0;j<nodes;j++)
						offspring.weight[j][i] = father.weight[j][i];
					offspring.bias[i] = father.bias[i];
				}
				//return offspring;
			}
		/*	else if (method == 2) {
				not yet implemented
			}*/
			else this.xOverWith(other, method);
		}
		
		return this;
	}
	
	public float activate (float x)
	{		
		if(x>=1)
			return 1;
		else if(x<=0)
			return 0;
		else 
			return (short)x;
		
	}
	
	public void reset()
	{
		for(int i=0;i<nodes;i++)
			this.output[i] = 0;
	}
	
	/** Provides the networks output for the given input */
	public Vector<Float> getOutput(Vector<Float> input) {
		for (int i=0;i<stepnumber-1;i++) {
			getStep(input);
		}
		return getStep (input);
	}

	
	public Vector<Float> getStep(Vector<Float> input) {
		if(input.size() != this.input_nodes)
			throw new IllegalArgumentException("Input vector size inappropriate!");
		
		for(int i = 0;i<input.size();i++) {
			this.output[i] = input.get(i);
		}	
		
		//propagate input to hidden layer
		for(int i = input_nodes ; i<nodes - output_nodes ; i++) //all hidden nodes
		{
			float sum = 0;
			for(int j = 0 ; j < input_nodes ; j++) //all input nodes
			{
				sum += this.weight[j][i] * this.output[j];
			}
			this.activation[i] = this.bias[i] + sum;
			this.output[i] = activate(this.activation[i]);
		}
		
		//propagate information to output layer
		
		for(int i = nodes-output_nodes;i<nodes;i++) //all output nodes
		{
			float sum = 0;
			for(int j = input_nodes ; j<nodes - output_nodes ; j++) //all hidden nodes
			{
				sum += this.weight[j][i] * this.output[j];
			}
			this.activation[i] = this.bias[i] + sum;
			this.output[i] = activate(this.activation[i]);
		}
				
		Vector<Float> outputVector = new Vector<Float>();
		
		for(int i = nodes-output_nodes;i<nodes;i++)
		{
			outputVector.add(this.output[i]);
		}
		//System.out.println (input.get(0)+","+input.get(1)+","+input.get(2));
		//System.out.println (outputVector.get(0));
		return outputVector;
	}

	@Override
	public void setRank(int rank) {
		this.rank=rank;	
	}

	@Override
	public double diffTo(IRepresentation o) {
		double diff=0;
		
		if(!(o instanceof ThreeLayerNetwork))
			throw new IllegalArgumentException("diffTo between different network classes not possible!");
		else
		{
			ThreeLayerNetwork network = (ThreeLayerNetwork) o;
			
			for(int i = 0;i<this.weight.length;i++ )
			{
				for(int j = 0;j<this.weight[i].length;j++)
				{
					diff += Math.abs(network.weight[i][j] - this.weight[i][j]); 
				}
			}
			for(int i =0;i<this.bias.length;i++)
				diff += Math.abs(network.bias[i] - this.bias[i]);
		}
		
		return diff;
	}

	public IRepresentation clone() {
		ThreeLayerNetwork res = new ThreeLayerNetwork();
		res.addProperties(getProperties());
		res.generate(input_nodes, output_nodes, generator);
		for(int i = 0;i<this.weight.length;i++ )
		{
			for(int j = 0;j<this.weight[i].length;j++)
			{
				res.weight[i][j] = this.weight[i][j]; 
			}
		}
		res.bias = this.bias.clone();
		res.output = this.output.clone();
		res.activation = this.activation.clone();
		res.rank = this.rank;
		return res;
	}
	
	public float getWeight(int nr, int from)
	{
		if(nr >= nodes)
			throw new IllegalArgumentException("Id "+nr+" is not a node!");
		{
			if(from >= nodes)
				throw new IllegalArgumentException("Id "+from+" is not a node!");
			else
				return this.weight[from][nr];
		}
	}
	
	/** Method used for saving the properties of this component*/
	public Element exportXml (Element element) {
		Element nn = element.addElement("ThreeLayerNetwork");
		nn.addAttribute("input_nodes",String.valueOf(this.input_nodes));
		nn.addAttribute("output_nodes",String.valueOf(this.output_nodes));
		nn.addAttribute("nodes",String.valueOf(this.nodes));
		nn.addAttribute("weight_range",String.valueOf(this.weight_range));
		nn.addAttribute("bias_range",String.valueOf(this.bias_range));
		
	/*	Element simulation = nn.addElement("simulation");
		simulation.addAttribute("iteration",String.valueOf(this.iteration));
		simulation.addAttribute("score",String.valueOf(this.score));*/
		
		Element dnodes = nn.addElement("nodes");
		
		for(int i = 0;i<this.nodes;i++)
		{
			Element node;
			if(i<input_nodes)
			{
				node = dnodes.addElement("node");
				node.addAttribute("nr", String.valueOf(i));
				node.addAttribute("type", "inputNode");
			}
			else
			{
				node = dnodes.addElement("node");
				node.addAttribute("nr", String.valueOf(i));
				node.addAttribute("type", "node");
				Element bias = node.addElement("bias");
				bias.addText(String.valueOf(this.bias[i]));
				
				Element weights = node.addElement("weights");
				for(int j = 0;j<this.nodes;j++)
				{
					Element weight = weights.addElement("weight");
					weight.addAttribute("from",String.valueOf(j));
					weight.addText(String.valueOf(this.getWeight(i, j)));
				}
			}
		}
				
		return nn;
	}

	@Override
	public IRepresentation loadFromXML(Node nd) {
		try {
			if (this.generator == null) this.generator = new NESRandom();
			this.input_nodes = Integer.parseInt(nd.valueOf("./@input_nodes"));
			this.output_nodes = Integer.parseInt(nd.valueOf("./@output_nodes"));
			this.nodes = Integer.parseInt(nd.valueOf("./@nodes"));
			this.weight_range = Float.parseFloat(nd.valueOf("./@weight_range"));
			this.bias_range = Float.parseFloat(nd.valueOf("./@bias_range"));
			//this.iteration = Integer.parseInt(nd.valueOf("./simulation/@iteration"));
			//this.score = Integer.parseInt(nd.valueOf("./simulation/@score"));
			this.activation = new float[this.nodes];
			this.output = new float[this.nodes];
			this.bias = new float[this.nodes];
			this.weight = new float[this.nodes][this.nodes];
			Node dnodes = nd.selectSingleNode("./nodes");
			for(int nr = this.input_nodes;nr<this.nodes;nr++)
			{
				Node curnode = dnodes.selectSingleNode("./node[@nr='"+nr+"']");
				if(curnode == null)
					throw new IllegalArgumentException("ThreeLayerNetwork: node tags inconsistent!" +
							"\ncheck 'nr' attributes and nodes count in nnetwork!");
				this.bias[nr] = Float.parseFloat(curnode.valueOf("./bias"));
				Node dweights = curnode.selectSingleNode("./weights");
				for(int from=0;from<this.nodes;from++)
				{
					String ws = dweights.valueOf("./weight[@from='"+from+"']");
					if(ws.length() == 0)
						throw new IllegalArgumentException("ThreeLayerNetwork: weight tags inconsistent!" +
								"\ncheck 'from' attributes and nodes count in nnetwork!");
					float val = Float.parseFloat(ws);
					this.weight[from][nr] = val;
				}
			}
			//this.gahist = new GAHistory(this.config, nd);
		}
		catch (NumberFormatException e)
		{
			throw new IllegalArgumentException("ThreeLayerNetwork: NumberFormatException! Check XML File");
		}
		return this;
	}

	@Override
	public Hashtable<String, String> getDetails() {
		Hashtable<String, String> result = new Hashtable<String, String>();
		result.put("input neurons", Integer.toString(this.input_nodes));
		result.put("output neurons", Integer.toString(this.output_nodes));
		result.put("hidden neurons", Integer.toString(nodes-input_nodes-output_nodes));
		result.put("weight range", Float.toString(this.weight_range));
		result.put("bias range", Float.toString(this.bias_range));
		for (int n = 0; n < nodes;n++) {
			result.put("bias node "+n, Float.toString(this.bias[n]));
			for(int from=0;from<this.nodes;from++) {
				result.put("weight from "+from+" to "+n, Float.toString(this.weight[from][n]));
			}
		}
		
		return result;
	}

	@Override
	public IRepresentation loadFromXML(Node nd, NESRandom r) {
		this.generator = r;
		return loadFromXML(nd);
	}

	@Override
	public void exportToFile(File saveFile) {
		System.out.println ("Exporting Pajek network file to "+saveFile.getName());
		try{
		    //Create file 
		    FileWriter fstream = new FileWriter(saveFile);
		    BufferedWriter out = new BufferedWriter(fstream);
		    out.write("*Vertices "+this.nodes);
		    out.newLine(); //use newline instead of \n to avoid system dependency
		    //input neurons
		    for (int i=1;i<input_nodes+1;i++) {
		    	out.write(i+" \"I"+i+"\"");
		    	out.newLine();
		    }
		    for (int h=input_nodes+1;h<nodes-output_nodes+1;h++) {
		    	out.write(h+" \"H"+(h-input_nodes)+"\"");
		    	out.newLine();
		    }
		    int a=1;
		    for (int o=nodes-output_nodes+1;o<nodes+1;o++) {
		    	out.write(o+" \"O"+a+"\"");
		    	out.newLine();
		    	a++;
		    }
		    
		    //Edges
		    out.write("*Edges");
		    out.newLine();
		    
		    //add edges from hidden neurons to input neurons
		    for (int n = input_nodes; n < (nodes-output_nodes);n++) { //take all hidden neurons
		    	for (int from = 0;from <input_nodes;from++) { //take all input neurons
		    		out.write((n+1)+" "+(from+1)+" "+weight[from][n]);
		    		out.newLine();
		    	}
			}
		    
		    //add edges from output neurons to hidden neurons
		    for (int n = (nodes-output_nodes); n < nodes;n++) { //take all output neurons
		    	for (int from = input_nodes;from <(nodes-output_nodes);from++) { //take all hidden neurons
		    		out.write((n+1)+" "+(from+1)+" "+weight[from][n]);
		    		out.newLine();
		    	}
		    }
		    
		    //Close the output stream
		    out.close();
		    } catch (Exception e){
		      System.err.println("Error: " + e.getMessage());
		    }
	}

}
