/*
 * Copyright (C) 2009 Istvan Fehervari, Wilfried Elmenreich
 * Original project page: http://www.frevotool.tk
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License Version 3 as published
 * by the Free Software Foundation http://www.gnu.org/licenses/gpl-3.0.txt
 *
 * There is no warranty for this free software. The GPL requires that 
 * modified versions be marked as changed, so that their problems will
 * not be attributed erroneously to authors of previous versions.
 */
package interfaces;

import java.util.Hashtable;
import java.util.Random;

/**
 * Provides a general interface to formulate own problems
 * 
 * @author Istvan Fehervari, istvan.fehervari@uni-klu.ac.at
 *
 */
public abstract class IProblem extends IComponent {
	
	protected Random generator;
	
	/**
	 * @return Use this random generator to have repeatable sessions
	 */
	public Random getRandom() {
		if (generator == null) {
			generator = new Random();
		}
		return generator;
	}

	/**
	 * @param generator the generator to set
	 */
	public void setRandom(Random generator) {
		this.generator = generator;
	}

	/** This function is called when the user request a visual replay from the problem. It is recommended to use advanced threading to display it like swingworkers */
	public abstract void replayWithVisualization(IRepresentation[] candidates);

	/** Returns the result of the simulation. Type depends on the size of the passed array.
	 * 
	 * @return The fitness value if size = 1. Higher value means better fitness.<br>
	 * <b>WARNING</b>: Use ONLY positive values or ONLY negative values (where the best solution is maxed at 0). 
	 * Mixing positive and negative values might make some ranking or method components to fail. 
	 * <br><br>
	 * else 1 if p1 wins, 2 if p2 wins, 0 if tie</pre> */
	public abstract double getResult(IRepresentation[] candidates);
	
	/** Override this method to get a more sophisticated output which might be needed for some simulation runs */
	public float[] getMoreResult(IRepresentation[] candidates){
		float[] res = new float [1];
		res[0] = (float) getResult(candidates);
		return res;
	}
	
	/** Returns the new requirements HashTable which is modified according to the properties. Override this method to allow requirement changes for your class*/
	public Hashtable<String, PropEntry> adjustRequirements(Hashtable<String, PropEntry> requirements, Hashtable<String, PropEntry> properties) {
		return requirements;
	}
	
}
