/*
 * Copyright (C) 2009 Istvan Fehervari, Wilfried Elmenreich
 * Original project page: http://www.frevotool.tk
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License Version 3 as published
 * by the Free Software Foundation http://www.gnu.org/licenses/gpl-3.0.txt
 *
 * There is no warranty for this free software. The GPL requires that 
 * modified versions be marked as changed, so that their problems will
 * not be attributed erroneously to authors of previous versions.
 */
package interfaces;

import java.io.File;
import java.util.Hashtable;
import java.util.Vector;

import org.dom4j.Element;
import org.dom4j.Node;

import utils.NESRandom;


public abstract class IRepresentation extends IComponent implements Comparable <IRepresentation>{

	private double fitness;
	
	/** Returns the score of this representation */
	public double getFitness() {
		return fitness;
	}

	/** Sets the score of this representation */
	public void setFitness(double fitness) {
		this.fitness = fitness;
	}
	
	public abstract void exportToFile(File savefile);

	/**
	 * Initiates a mutation on this representation
	 * @param severity (%) Should be between 0-100, interpreted by the representation.
	 * @param probability (%) Should be between 0-100, probability for one genome to be mutated.
	 * @param method Selected mutation method (if more). Use getNumberofMethods() to obtain the number of implemented methods.
	 */
	public abstract IRepresentation mutate(float severity, int probability, int method);
	
	/** Initiates a crossover on this representation and with the given representation in the given method. Use getNumberofXoverMethods() to obtain the number of usable methods */
	public abstract IRepresentation xOverWith(IRepresentation other, int method);
	
	/** Returns the number of implemented mutation methods */
	public abstract int getNumberofMutationMethods();
	
	/** Returns the number of implemented crossover methods */
	public abstract int getNumberofXoverMethods();
	
	/** Generates a given instance of the representation */
	public abstract void generate(int inputnumber,int outputnumber, NESRandom random);
	
	/** Returns the output of the representation for the given input. The output must be within the range of 0 and 1 */
	public abstract Vector<Float> getOutput(Vector<Float> input);
	
	/** Resets the representation (erases cache)*/
	public abstract void reset();

	/** Sets the rank of the representation */
	public abstract void setRank(int i);

	/** Compares this representation to another one. Useful for measuring diversity */
	public abstract double diffTo(IRepresentation representation);
	
	/** Returns a clone of this representation */
	public abstract IRepresentation clone();
	
	/** Method used for saving the properties of this component*/
	public abstract Element exportXml (Element element);
	
	/** Loads this representation from an xml file */
	public abstract IRepresentation loadFromXML(Node nd);
	
	/** Loads this representation from an xml file with the given random object */
	public abstract IRepresentation loadFromXML(Node nd, NESRandom r);
	
	/** Returns a human-readable table (keys-values) about the details/configuration of this specific instance*/
	public abstract Hashtable<String,String> getDetails();
	
	/** Returns a unique name of this representation calculated by its parameters/structure. This is mostly used for identification. */
	public abstract String getUniqueName();
	
	public int compareTo(IRepresentation other) {
		if (this.getFitness() > other.getFitness()) return 1; //our score is larger
		else if (this.getFitness() < other.getFitness()) return -1; //other one is larger
		return 0;
	}

}
