/*
 * Copyright (C) 2009 Istvan Fehervari, Wilfried Elmenreich
 * Original project page: http://www.frevotool.tk
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License Version 3 as published
 * by the Free Software Foundation http://www.gnu.org/licenses/gpl-3.0.txt
 *
 * There is no warranty for this free software. The GPL requires that 
 * modified versions be marked as changed, so that their problems will
 * not be attributed erroneously to authors of pious versions.
 */
package main;
import graphics.FrevoTheme;
import graphics.FrevoWindow;
import interfaces.ComponentXMLdata;
import interfaces.IComponent;
import interfaces.IMethod;
import interfaces.IProblem;
import interfaces.IRepresentation;
import interfaces.PropEntry;
import interfaces.PropTypes;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Scanner;
import java.util.Vector;

import javax.swing.UIManager;
import javax.swing.plaf.metal.MetalLookAndFeel;

import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.Node;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.XMLWriter;

import utils.NESRandom;
import utils.SafeSAX;
import utils.StatKeeper;

/** This is the core class of FREVO. It handles the components and operates the simulation run */
public class FrevoMain {
	
	/** Indicates the current debug level */
	public static int DEBUGLEVEL = 0x00;
	/** Is true when FREVO is started in graphical mode */
	public static boolean graphics;
	/** This object holds the reference of the graphical component of FREVO */
	public static FrevoWindow mainWindow;
	/** Holds the root directory of the program*/
	public final static String currentDir = System.getProperty("user.dir");
	/** The current major version */
	public final static String MAJORVERSION = "0.8";
	/** The current minor version */
	public static String MINORVERSION;
	//----------------------static fields for referencing-------------------
	/** Static reference for indexing problem components */
	public static final int FREVO_IPROBLEM = 0;
	/** Static reference for indexing method components */
	public static final int FREVO_IMETHOD = 1;
	/** Static reference for indexing representation components */
	public static final int FREVO_IREPRESENTATION = 2;
	/** Static reference for indexing ranking components */
	public static final int FREVO_IRANKING = 3;
	
	/** File extension for frevo session files */
	public static final String FREVO_SESSION_EXTENSION = "zse";
	/** File extension for frevo result files */
	public static final String FREVO_RESULT_EXTENSION = "zre";
	
	/** If true console output will be written directly to a file */
	public static final boolean FREVO_REDIRECTCONSOLE_TO_FILE = false;
	
	/** A static object holding the results to keep them in memory */
	private static Object results;
	/** The current simulation run out of <i>maxRun</i>*/
	public static int currentRun = 1;
	/** The number of total simulation runs */
	public static int maxRun = 1;
	/** Starting seed of the simulation run */
	public static long SEED = 12345;
	/** Custom name of the session, mainly used for saving */
	public static String customName = "DEFAULT";
	/** Statistics object, the first elements should be the index */
	public static ArrayList<StatKeeper> STATISTICS;
	
	/** Startup parameters for FREVO */
	public static String[] PARAMETERS;
	
	/** Array containing the selection of different components stored in arraylists. use FREVO_ICOMPONENT values to access 
	 * Enter value -1 if not selected */
	public static int[] SELECTED_COMPONENT = new int[]{-1,-1,-1,-1};
	
	/** FREVO properties read from file on startup */
	private static ArrayList<TagElement> problemTagList;
	private static ArrayList<TagElement> methodTagList;
	private static ArrayList<TagElement> representationTagList;
	private static ArrayList<TagElement> rankingTagList;
	
	public static int[] componentBrowserParameters = new int[]{600,600,273,271};
	public static int[] mainWindowParameters = new int[]{600,600};
	
	//Component storage
	public static ArrayList<ComponentXMLdata> iProblems = new ArrayList<ComponentXMLdata>(); //Storing IProblems
	public static ArrayList<ComponentXMLdata> iMethods = new ArrayList<ComponentXMLdata>(); //Storing IMethods
	public static ArrayList<ComponentXMLdata> iRepresentations = new ArrayList<ComponentXMLdata>(); //Storing IRepresentations
	public static ArrayList<ComponentXMLdata> iRankings = new ArrayList<ComponentXMLdata>(); //Storing IRankings

	//Flags for FREVO running state
	public static boolean isLaunchable = false;
	public static boolean isRunning = false;
	
	public static void main(String[] args) {
		
		PARAMETERS = args;
		MINORVERSION = "$Rev: 265 $".substring(6,8);
		System.out.println("Launching FREVO "+"v"+MAJORVERSION+" Rev:"+MINORVERSION);
		
		// Initialization
		STATISTICS = new ArrayList<StatKeeper>(); //reset statistics object
     
        // Execution
		if (args.length == 0) { //graphical version
			FrevoMain.graphics = true;
			
			//launch GUI window
			java.awt.EventQueue.invokeLater(new Runnable() {
	            public void run() {
	            	try {
	            		//use standard look and feel
	            		MetalLookAndFeel.setCurrentTheme(new FrevoTheme());
	            		UIManager.setLookAndFeel(new MetalLookAndFeel());
	            		
	            		loadComponents(); //Load components to the proper array
	                    loadProperties(); //Load FREVO properties
	                    
	                    for (int i = 0;i<4;i++) SELECTED_COMPONENT[i] = -1; //Deselect components
	                    
						mainWindow = new FrevoWindow();
						mainWindow.setVisible(true);
					} catch (Exception e) {
						e.printStackTrace();
						//no graphics available!
						System.out.println ("FREVO: graphics are missing!\nTry \'frevo -?\' for more information");
					}
	            }
            });
		}
		else {
			if (args[0].equals("-s")) { //Load session and run
				if (FREVO_REDIRECTCONSOLE_TO_FILE) {
					try {
						
						System.out.println ("Console output is written to log.txt");
						File logfile = new File("log.txt");
						OutputStream output = new FileOutputStream(logfile);
						PrintStream printOut = new PrintStream(output);
						System.setOut(printOut);
					} catch (FileNotFoundException e) {
						e.printStackTrace();
					}
				}
				loadComponents(); //Load components to the proper array
                loadProperties(); //Load FREVO properties
                
                for (int i = 0;i<4;i++) SELECTED_COMPONENT[i] = -1; //Deselect components
                
				runSessionFromFile(args[1]);
				
			}
			else if (args[0].equals("-e")) {//runs (evaluates) best solution only
				System.out.println ("Evaluating best representation from "+args[1]);
				if (FREVO_REDIRECTCONSOLE_TO_FILE) {
					try {						
						System.out.println ("Console output is written to log.txt");
						File logfile = new File("log.txt");
						OutputStream output = new FileOutputStream(logfile);
						PrintStream printOut = new PrintStream(output);
						System.setOut(printOut);
					} catch (FileNotFoundException e) {
						e.printStackTrace();
					}
				}
				loadComponents(); //Load components to the proper array
                loadProperties(); //Load FREVO properties
                
                for (int i = 0;i<4;i++) SELECTED_COMPONENT[i] = -1; //Deselect components
                
				evaluateResult(args[1]);
                //runSessionFromFile(args[1]);
			}			
			else if (args[0].equals("-?") || args[0].equals("?") || args[0].equals("/?") ) { //Load session and run
				printUsage ();
			} else {
				System.out.println ("ERROR: wrong parameter set!");
				printUsage();
			}
		}

	}
	
	/** Prints the usage parameters to the console */
	private static void printUsage () {
		System.out.println ("USAGE:");
		System.out.println ("-s FILENAME : loads and executes the session file, eg. -s test.zse");
	}
	
	/** Loads component data from all configuration files */
 	private static void loadComponents() {
		String[] names = new String[]{"Problems","Methods","Representations","Rankings"};
		
		for (int ctypes = 0; ctypes < SELECTED_COMPONENT.length; ctypes++) {
			File iComponentdir = new File (currentDir+"//Components//"+names[ctypes]);
			
			File[] files = iComponentdir.listFiles();    
		    FilenameFilter fileFilter = new FilenameFilter() { 	    // This filter only returns .xml files
		        public boolean accept(File file, String name) {
                    if (name.length() < 4) return false;
		        	if ((name.substring(name.length()-4, name.length())).equals(".xml")) return true;
		        	else return false;
		        }
		    };
		    files = iComponentdir.listFiles(fileFilter); //we have all i* XMLs here
		    
		    for (int i = 0; i<files.length;i++) {
		    	if (files.length != 0) 
		    	try {
		    		ComponentXMLdata cdata = readIComponentFromXml(files[i].getAbsolutePath(),false,ctypes); //should load requirement fields as well
		    		if (cdata != null) getComponentList(ctypes).add(cdata);	    		
				} catch (RemoteException e) {
					e.printStackTrace();
				} 
		    }
		}
		int total = (iProblems.size()+iMethods.size()+iRepresentations.size()+iRankings.size()  );
	        
	    System.out.println("Total components loaded: "+total+ " (P:"+iProblems.size()+",M:"+iMethods.size()+",Re:"+iRepresentations.size()+",Ra:"+iRankings.size()+")");
	}
	
	@SuppressWarnings("unchecked")
	public static ComponentXMLdata readIComponentFromXml(String filename, boolean validateXml, int cType) throws RemoteException {
        boolean error = false;
        ComponentXMLdata result = new ComponentXMLdata(cType);
        
        File file = new File(filename);
        Document doc = SafeSAX.read(file);

        try {
        	//Load configuration section
            Node nd = doc.selectSingleNode("/icomponent/config");
            if (nd == null) {
                System.out.println("No config tag found - Excluding this IProblem");

            } else {
                List<?> npops = nd.selectNodes(".//entry");
        		Iterator<?> it = npops.iterator();

        		while(it.hasNext()) {
        			Element el = (Element) it.next();
        			String key = el.valueOf("./@key");
        			String value = el.valueOf("./@value");
        			
        			if (key.equals("classdir")) {
        				result.setClassDir(value);
        			}
        			else if (key.equals("classname")) {
        				result.setClassName(value);
        			}
        			else if (key.equals("name")) {
        				result.setName(value);
        			}
        			else if (key.equals("description")) {
        				result.setDescription(value);
        			}
        			else if (key.equals("image")) {
        				result.setImageLocation(value);
        			}
        			else if (key.equals("tags")) {
        				ArrayList<String> taglist = new ArrayList<String>();
        				Scanner sc = new Scanner(value);
        				sc.useDelimiter(",");
        				while (sc.hasNext()) taglist.add(sc.next());
        				result.setTags(taglist);
        			}
        		}
        		
        		//Load class file
        		//file = new File(currentDir+"//Components//Problems//"+result.getClassDir());
		        file = new File(currentDir);
        		// Convert File to a URL
		        URL url;
				try {
					url = file.toURI().toURL();
					URL[] urls = new URL[]{url};
					ClassLoader cl = new URLClassLoader(urls);
					Class<IComponent> cls = (Class<IComponent>)(cl.loadClass(result.getClassName()));
					result.setComponentClass(cls);
				}
				catch (Exception e) {
					e.printStackTrace();
				}
            }
            //Load properties section    
            Node nd2 = doc.selectSingleNode("/icomponent/properties");
            if (nd2 == null) {
                System.out.println("No properties tag found");

            } else {
                List<?> npops = nd2.selectNodes(".//entry");
        		Iterator<?> it = npops.iterator();

        		//create hashmap
        		Hashtable<String, PropEntry> proptable = new Hashtable<String, PropEntry>();
        		
        		while(it.hasNext()) {
        			Element el = (Element) it.next();
        			String key = el.valueOf("./@key");
        			PropTypes type = PropTypes.valueOf(el.valueOf("./@type"));
        			String value = el.valueOf("./@value");
        			
        			if(!checkType(type, value))
        				throw new IllegalArgumentException("Value \""+value+"\" for Key \""+
        						key+"\" is not of type "+type+"!");
        			proptable.put(key, new PropEntry(value,type));
        			
        		}
        		//Add hashmap to the object
        		result.setProperties(proptable);
            }
            //Load requirements
            if (cType == FREVO_IPROBLEM) {
            	Node nd3 = doc.selectSingleNode("/icomponent/requirements");
                if (nd3 == null) {
                    System.err.println("No requirements tag found in IProblem!");

                } else {
                    List<?> npops = nd3.selectNodes(".//entry");
            		Iterator<?> it = npops.iterator();

            		//create hashmap
            		Hashtable<String, PropEntry> reqtable = new Hashtable<String, PropEntry>();
            		
            		while(it.hasNext()) {
            			Element el = (Element) it.next();
            			String key = el.valueOf("./@key");
            			PropTypes type = PropTypes.valueOf(el.valueOf("./@type"));
            			String value = el.valueOf("./@value");
            			
            			if(!checkType(type, value))
            				throw new IllegalArgumentException("Value \""+value+"\" for Key \""+
            						key+"\" is not of type "+type+"!");
            			reqtable.put(key, new PropEntry(value,type));
            			
            		}
            		//Add hashmap to the object
            		result.setRequirements(reqtable);
                }
            	
            }
        } catch (OutOfMemoryError mem) {
            System.err.println("Could not import! (Out of memory)");
            error = true;
        } catch (IllegalArgumentException e) {
            System.out.println("IllegalArgumentException\n" + e.getMessage());
            error = true;
        } catch (Exception e) {
        	System.out.println ("Error while importing");
        	error = true;
        }
        if (!error) {
         //   System.out.println("Successfully Completed!");
        } else {
            System.err.println("Error while importing!");
        }
        return result;
    }
	
	public static ComponentXMLdata getComponent(String classname, int type) {
		ArrayList<ComponentXMLdata> list = getComponentList(type);
		ComponentXMLdata result = null;
		for (int i = 0; i< list.size();i++) {
			if (list.get(i).getClassName().equals(classname) ) result = list.get(i);
		}
		return result;
	}
	
	public static Document loadSession(File loadFile) {
		//loading session
		Document doc = null;
        boolean error = false;
                
        doc = SafeSAX.read(loadFile);
        
        try {
        	//Load configuration section
            Node nd = doc.selectSingleNode("/frevo/sessionconfig");
            if (nd == null) {
               // System.out.println("No config tag found - Excluding configuration");

            } else { //load configuration
                List<?> npops = nd.selectNodes(".//configentry");
        		Iterator<?> it = npops.iterator();

        		while(it.hasNext()) {
        			Element el = (Element) it.next();
        			String key = el.valueOf("./@key");
        			String value = el.valueOf("./@value");
        			
        			if (key.equals("CustomName")) {
        				FrevoMain.customName = value;
        			}
        			else if (key.equals("NumberofRuns")) {
        				FrevoMain.maxRun = Integer.parseInt(value);
        			}
        			else if (key.equals("StartingSeed")) {
        				FrevoMain.SEED = Long.parseLong(value);
        			}
        			
        		}				        		
            }// config
        
            //Load problem
            Node problemnode = doc.selectSingleNode("/frevo/problem");
            //select appropriate component 
            String classname = problemnode.valueOf("./@class");
            ComponentXMLdata problem = getComponent(classname,FREVO_IPROBLEM);
            if (problem == null) throw new Error ("No such component found: "+classname);
            else {
            	SELECTED_COMPONENT[FREVO_IPROBLEM] = iProblems.indexOf(problem);
            	//load configuration
	            List<?> pentries = problemnode.selectNodes(".//problementry");
	        	Iterator<?> it = pentries.iterator();

	        	while(it.hasNext()) {
	        		Element el = (Element) it.next();
	        		String key = el.valueOf("./@key");
	        		String value = el.valueOf("./@value");
	        		PropEntry pe = problem.getProperties().get(key);	
	        		pe.setValue(value);
	        	}
            }
            	
            //Load method
            Node methodnode = doc.selectSingleNode("/frevo/method");
            //select appropriate component 
            classname = methodnode.valueOf("./@class");
            ComponentXMLdata method = getComponent(classname,FREVO_IMETHOD);
            if (method == null) {
            	if (classname.equals("")) System.err.println ("No method to load");
            	else throw new Error ("No such component found: "+classname);
            }
            else {
            	SELECTED_COMPONENT[FREVO_IMETHOD] = iMethods.indexOf(method);
            	//load configuration
	            List<?> pentries = methodnode.selectNodes(".//methodentry");
	        	Iterator<?> it = pentries.iterator();

	        	while(it.hasNext()) {
	        		Element el = (Element) it.next();
	        		String key = el.valueOf("./@key");
	        		String value = el.valueOf("./@value");
	        		PropEntry pe = method.getProperties().get(key);	
	        		pe.setValue(value);
	        	}
            }
            
            //Load representation
            Node representationnode = doc.selectSingleNode("/frevo/representation");
            //select appropriate component 
            classname = representationnode.valueOf("./@class");
            ComponentXMLdata representation = getComponent(classname,FREVO_IREPRESENTATION);
            if (representation == null) { 
            	if (classname.equals("")) System.err.println ("No representation to load");
            	else throw new Error ("No such component found: "+classname);
            }
            else {
            	SELECTED_COMPONENT[FREVO_IREPRESENTATION] = iRepresentations.indexOf(representation);
            	//load configuration
	            List<?> pentries = representationnode.selectNodes(".//representationentry");
	        	Iterator<?> it = pentries.iterator();

	        	while(it.hasNext()) {
	        		Element el = (Element) it.next();
	        		String key = el.valueOf("./@key");
	        		String value = el.valueOf("./@value");
	        		PropEntry pe = representation.getProperties().get(key);	
	        		pe.setValue(value);
	        	}
            }
            
            //Load ranking
            Node rankingnode = doc.selectSingleNode("/frevo/ranking");
            //select appropriate component 
            classname = rankingnode.valueOf("./@class");
            ComponentXMLdata ranking = getComponent(classname,FREVO_IRANKING);
            if (ranking == null) {
            	if (classname.equals("")) System.err.println ("No ranking to load");
            	else throw new Error ("No such component found: "+classname);
            }
            else {
            	SELECTED_COMPONENT[FREVO_IRANKING] = iRankings.indexOf(ranking);
            	//load configuration
	            List<?> pentries = rankingnode.selectNodes(".//rankingentry");
	        	Iterator<?> it = pentries.iterator();

	        	while(it.hasNext()) {
	        		Element el = (Element) it.next();
	        		String key = el.valueOf("./@key");
	        		String value = el.valueOf("./@value");
	        		PropEntry pe = ranking.getProperties().get(key);	
	        		pe.setValue(value);
	        	}
            }
            
        } catch (OutOfMemoryError mem) {
            System.err.println("Could not import! (Out of memory)");
            error = true;
        } catch (IllegalArgumentException e) {
            System.out.println("IllegalArgumentException\n" + e.getMessage());
            error = true;
        }
        if (!error) {
         //   System.out.println("Successfully Completed!");
        } else {
            System.err.println("Error while importing!");
        }
        return doc;
	}

	public static void continueSession(String name, NESRandom random) {
		try {
			//Instantiate components
			final IMethod method = (IMethod)(iMethods.get(SELECTED_COMPONENT[FREVO_IMETHOD])).prepareComponent();
			
			//Feed components to method instance
			method.addIProblem(iProblems.get(SELECTED_COMPONENT[FREVO_IPROBLEM]));
			method.addIRepresentation(iRepresentations.get(SELECTED_COMPONENT[FREVO_IREPRESENTATION]));
			method.addIRanking(iRankings.get(SELECTED_COMPONENT[FREVO_IRANKING]));
			
			method.setRandom(random); //to running parameters
			method.addoutputname(name);
			
			System.out.println ("Continuing simulation");
			mainWindow.startSim(method);

								
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		
		
	}
	
	/** Starts a simulation with the given name and starting seed */
	public static void runSimulation (String name, Long seed) {
		publishResults(null); //clear results cache
		try {
			//Instantiate components
			final IMethod method = (IMethod)(iMethods.get(SELECTED_COMPONENT[FREVO_IMETHOD])).prepareComponent();
			
			//Feed components to method instance
			method.addIProblem(iProblems.get(SELECTED_COMPONENT[FREVO_IPROBLEM]));
			method.addIRepresentation(iRepresentations.get(SELECTED_COMPONENT[FREVO_IREPRESENTATION]));
			method.addIRanking(iRankings.get(SELECTED_COMPONENT[FREVO_IRANKING]));
			method.setRandomSeed(seed);
			
			method.addoutputname(name);
			
			System.out.println ("Starting simulation");
			if (graphics) {				
				mainWindow.startSim(method);
			}
			else {
				method.run(); //no threading
			}
								
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		
	}
	
	/** Loads FREVO properties */
	private static void loadProperties() {
		try {
			Document doc = null;             
			doc = SafeSAX.read(new File("frevo_config.xml"));
			
			//load tags
			problemTagList = new ArrayList<TagElement>();
			methodTagList = new ArrayList<TagElement>();
			representationTagList = new ArrayList<TagElement>();
			rankingTagList = new ArrayList<TagElement>();
			
			Node tagcategories = doc.selectSingleNode("/frevo/tags");
			
			//problem tags
			List<?> items = tagcategories.selectNodes(".//problemtag");
			Iterator<?> it = items.iterator();

			while(it.hasNext()) {
				Element el = (Element) it.next();
				String name = el.valueOf("./@name");
				String tags = el.valueOf("./@tags");
				String imagename = el.valueOf("./@image");
				
				ArrayList<String> taglist = new ArrayList<String>();
				Scanner sc = new Scanner(tags);
				sc.useDelimiter(",");
				while (sc.hasNext()) taglist.add(sc.next());
				
				problemTagList.add(new TagElement(name,taglist,imagename));
			}
			
			//method tags
			items = tagcategories.selectNodes(".//methodtag");
			it = items.iterator();

			while(it.hasNext()) {
				Element el = (Element) it.next();
				String name = el.valueOf("./@name");
				String tags = el.valueOf("./@tags");
				String imagename = el.valueOf("./@image");
				
				ArrayList<String> taglist = new ArrayList<String>();
				Scanner sc = new Scanner(tags);
				sc.useDelimiter(",");
				while (sc.hasNext()) taglist.add(sc.next());
				
				methodTagList.add(new TagElement(name,taglist,imagename));
			}
			
			//representation tags
			items = tagcategories.selectNodes(".//representationtag");
			it = items.iterator();

			while(it.hasNext()) {
				Element el = (Element) it.next();
				String name = el.valueOf("./@name");
				String tags = el.valueOf("./@tags");
				String imagename = el.valueOf("./@image");
				
				ArrayList<String> taglist = new ArrayList<String>();
				Scanner sc = new Scanner(tags);
				sc.useDelimiter(",");
				while (sc.hasNext()) taglist.add(sc.next());
				
				representationTagList.add(new TagElement(name,taglist,imagename));
			}
			
			//ranking tags
			items = tagcategories.selectNodes(".//rankingtag");
			it = items.iterator();

			while(it.hasNext()) {
				Element el = (Element) it.next();
				String name = el.valueOf("./@name");
				String tags = el.valueOf("./@tags");
				String imagename = el.valueOf("./@image");
				
				ArrayList<String> taglist = new ArrayList<String>();
				Scanner sc = new Scanner(tags);
				sc.useDelimiter(",");
				while (sc.hasNext()) taglist.add(sc.next());
				
				rankingTagList.add(new TagElement(name,taglist,imagename));
			}
			
			//load preferred window size
			Node windowsizes =  doc.selectSingleNode("/frevo/window-sizes");
			
			items = windowsizes.selectNodes(".//window");
			it = items.iterator();
			
			while(it.hasNext()) {
				Element el = (Element) it.next();
				String name = el.valueOf("./@name");
				
				if (name.equals("main")) {
					String width = el.valueOf("./@width");
					String height = el.valueOf("./@height");
					mainWindowParameters[0] = Integer.parseInt(width);
					mainWindowParameters[1] = Integer.parseInt(height);
				} else	if (name.equals("componentbrowser")) {
					String width = el.valueOf("./@width");
					String height = el.valueOf("./@height");
					String topsplit = el.valueOf("./@topsplit");
					String bigsplit = el.valueOf("./@bigsplit");
					componentBrowserParameters[0] = Integer.parseInt(width);
					componentBrowserParameters[1] = Integer.parseInt(height);
					componentBrowserParameters[2] = Integer.parseInt(topsplit);
					componentBrowserParameters[3] = Integer.parseInt(bigsplit);
				}	
			}
		} catch (Exception e) {
			System.err.println ("No config file found, using defaults");
		}
		
		
	}
	
	public static class TagElement {
		private String name;
		private ArrayList<String> tags;
		private String imagename;
		private int elements = -1;
		
		public TagElement(String name,ArrayList<String> tags, String imagename) {
			this.name = name;
			this.tags = tags;
			this.imagename = imagename;
		}
		
		public String getName() {
			return name;
		}
		
		public ArrayList<String> getTags() {
			return tags;
		}
		
		public String getImageName() {
			return imagename;
		}
		
		public void setElements (int e) {
			this.elements = e;
		}
		
		public String toString() {
			if (elements != -1) return name+" ("+elements+")";
			else return name;
		}
		
	}
	
	public static ArrayList<TagElement> getCategories(int id) {
		switch (id) {
		case FREVO_IPROBLEM: return problemTagList;
		case FREVO_IMETHOD: return methodTagList;
		case FREVO_IREPRESENTATION: return representationTagList;
		case FREVO_IRANKING: return rankingTagList;
		}
		return null;
	}
	
	/** Starts a saving procedure calling the method's own saveResults function, the run-specific number will be added to the end of the file */
	public static void saveResult(String name, IMethod method) {
		//System.err.println ("Saving name: <"+name+">");
		Document doc = DocumentHelper.createDocument();
        doc.addDocType("frevo", null, System.getProperty("user.dir")+"//Components//ISave.dtd");
        Element dfrevo = doc.addElement("frevo");
        
        try {
            //export problem
        	Element problemsettings = dfrevo.addElement("problem");
        	ComponentXMLdata problem = FrevoMain.iProblems.get(FrevoMain.SELECTED_COMPONENT[FrevoMain.FREVO_IPROBLEM]);
        	problemsettings.addAttribute("class", problem.getClassName());
        	Vector<String> keys =  new Vector<String>(problem.getProperties().keySet());
        	for(String k:keys)
        	{
        		Element entry = problemsettings.addElement("problementry");
        		entry.addAttribute("key", k);
        		entry.addAttribute("type",problem.getType(k).toString());
        		entry.addAttribute("value", problem.getValue(k));
        	}
        	
        	//export method
        	Element methodsettings = dfrevo.addElement("method");
        	methodsettings.addAttribute("class", method.getXMLData().getClassName());
        	keys =  new Vector<String>(method.getProperties().keySet());
        	for(String k:keys)
        	{
        		Element entry = methodsettings.addElement("methodentry");
        		entry.addAttribute("key", k);
        		entry.addAttribute("type",method.getType(k).toString());
        		entry.addAttribute("value", method.getValue(k));
        	}
        	
        	//export ranking
        	Element rankingsettings = dfrevo.addElement("ranking");
        	ComponentXMLdata ranking = FrevoMain.iRankings.get(FrevoMain.SELECTED_COMPONENT[FrevoMain.FREVO_IRANKING]);
        	rankingsettings.addAttribute("class", ranking.getClassName());
        	keys =  new Vector<String>(ranking.getProperties().keySet());
        	for(String k:keys)
        	{
        		Element entry = rankingsettings.addElement("rankingentry");
        		entry.addAttribute("key", k);
        		entry.addAttribute("type",ranking.getType(k).toString());
        		entry.addAttribute("value", ranking.getValue(k));
        	}
        	
            //export representation
        	Element repsettings = dfrevo.addElement("representation");
        	ComponentXMLdata representation = FrevoMain.iRepresentations.get(FrevoMain.SELECTED_COMPONENT[FrevoMain.FREVO_IREPRESENTATION]);
        	repsettings.addAttribute("class", representation.getClassName());
        	keys =  new Vector<String>(representation.getProperties().keySet());
        	for(String k:keys)
        	{
        		Element entry = repsettings.addElement("representationentry");
        		entry.addAttribute("key", k);
        		entry.addAttribute("type",representation.getType(k).toString());
        		entry.addAttribute("value", representation.getValue(k));
        	}
            
        	// call method's own save solution
        	method.saveResults(dfrevo);

    		//save contents to the file
        	String runNumber = "";
        	if (maxRun > 1) {
        		runNumber = "_"+Integer.toString(currentRun);
        	}
            File file = new File(name+runNumber+".zre");

            OutputFormat format = OutputFormat.createPrettyPrint();
            format.setLineSeparator(System.getProperty("line.separator"));

            file.createNewFile();
            FileWriter out = new FileWriter(file);
            BufferedWriter bw = new BufferedWriter(out);
            XMLWriter wr = new XMLWriter(bw, format);
            wr.write(doc);
            wr.close();
            System.out.println("XML Writing Completed: "+name+runNumber+".zre");
        } catch (OutOfMemoryError mem) {
            System.err.println("Could not export! (Out of memory)");
        } catch (IOException e) {
            System.err.println("IOException while writing to XML!");
            e.printStackTrace();
        }
	}
	
	public static void publishResults(Object results) {
		FrevoMain.results = results;
	}
	
	public static Object getResults() {
		return FrevoMain.results;
	}
	
	public static ComponentXMLdata getComponent(int componentType, int componentNumber) {
		if (componentType == FREVO_IPROBLEM) return iProblems.get(componentNumber);
		else if (componentType == FREVO_IMETHOD) return iMethods.get(componentNumber);
		else if (componentType == FREVO_IREPRESENTATION) return iRepresentations.get(componentNumber);
		else if (componentType == FREVO_IRANKING) return iRankings.get(componentNumber);
		else throw new Error ("Invalid component type!");
	}
	
	public static ArrayList<ComponentXMLdata> getComponentList(int componentType) {
		if (componentType == FREVO_IPROBLEM) return iProblems;
		else if (componentType == FREVO_IMETHOD) return iMethods;
		else if (componentType == FREVO_IREPRESENTATION) return iRepresentations;
		else if (componentType == FREVO_IRANKING) return iRankings;
		else throw new Error ("Invalid component type!");
	}
	
	public static String getComponentDirectory(int componentType) {
		if (componentType == FREVO_IPROBLEM) return currentDir+"//Components//Problems//";
		else if (componentType == FREVO_IMETHOD) return currentDir+"//Components//Methods//";
		else if (componentType == FREVO_IREPRESENTATION) return currentDir+"//Components//Representations//";
		else if (componentType == FREVO_IRANKING) return currentDir+"//Components//Rankings//";
		else throw new Error ("Invalid component type!");
	}
	
	/** Sets the simulation launchable if all components are selected*/
	public static boolean checkState() {
		if ((SELECTED_COMPONENT[FREVO_IPROBLEM] != -1) && (SELECTED_COMPONENT[FREVO_IMETHOD] != -1) && (SELECTED_COMPONENT[FREVO_IRANKING] != -1) && (SELECTED_COMPONENT[FREVO_IREPRESENTATION] != -1) ){
			return true;
		}
		else return false;
	}
	
	public static boolean checkType(PropTypes type, String value) {
		try
		{
			switch (type)
			{
			case INT:
				Integer.parseInt(value);
				break;
			case LONG:
				Long.parseLong(value);
				break;
			case FLOAT:
				Float.parseFloat(value);
				break;
			case BOOLEAN:
				if ((value.equalsIgnoreCase("true")) || (value.equalsIgnoreCase("false"))) return true;
				else return false;
			//	break;
			}
		}
		catch (NumberFormatException e)
		{
			return false;
		}
		return true;
	}
	
	/** Returns the extension of a file */
	public static String getExtension(File f) {
		String fileName = f.getName();
	    int mid= fileName.lastIndexOf(".");
	    String ext=fileName.substring(mid+1,fileName.length());  
	    return (ext);
	}

	private static void runSessionFromFile(String xmlfile) {
		loadSession(new File(xmlfile));
		FrevoMain.currentRun = 0;
		FrevoMain.maxRun = 1;
		long seed = 12346454;
		runSimulation(FrevoMain.customName,seed);
	}
	
	private static void evaluateResult(String xmlfile) {
		try {
			for (int counter = 0; counter < 200; counter++) {
				IRepresentation bestnet = getRepresentation(new File(xmlfile),
						counter);//getBestNetwork(new File(xmlfile));
				ComponentXMLdata problem = iProblems
						.get(SELECTED_COMPONENT[FREVO_IPROBLEM]);
				//problem.getRequirements();
				IProblem p = (IProblem) (problem.prepareComponent());
				//p.setRandom();
				bestnet.reset(); //wipe clean state of the candidate 
				double fitness = p.getResult(new IRepresentation[] { bestnet });
				System.out.println("#"+counter+" fitness= " + fitness);
			}
			/*IRepresentation bestnet = getRepresentation(new File(xmlfile),2);
			NEAT n = (NEAT) bestnet;
			n.exportToPajek("errornet.net");*/
		} catch (Exception e) {
			e.printStackTrace();
		} 
		
	}
	
	@SuppressWarnings("unchecked")
	public static IRepresentation getRepresentation (File loadfile, int rank) {
		loadSession(loadfile);

		ComponentXMLdata representation =  iRepresentations.get(SELECTED_COMPONENT[FREVO_IREPRESENTATION]);
		IRepresentation net2h = null;
		
		Document doc = SafeSAX.read(loadfile);

		try {
			/*Node dpopulations = doc.selectSingleNode("/frevo/populations");
			List<?> npops = dpopulations.selectNodes(".//population");
			Iterator<?> it = npops.iterator();
			while(it.hasNext())
			{
				Node nd = (Node) it.next();				
				List<?> npops1 = nd.selectNodes("./*");
				Iterator<?> it1 = npops1.iterator();
					
				Node net = (Node) it1.next();
				net2h = (IRepresentation)(representation.prepareComponent());
				net2h.loadFromXML (net);					
				return net2h.clone();
			}*/
			Node nd = doc.selectSingleNode("/frevo/population");
			List<Node> npops1 = nd.selectNodes("./*");
			
			Node net = (Node) npops1.get(rank);
			net2h = (IRepresentation)(representation.prepareComponent());
			net2h.loadFromXML (net);					
			return net2h.clone();
			
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	/** Return the best candidate from the given result file */
	public static IRepresentation getBestNetwork (File loadfile) {
		return getRepresentation(loadfile, 0);
	}

	public static void updateStats() {
    	//generate folder if it doesnt exist
		String dirname = FrevoMain.customName+"_stats";
		(new File("./Results/"+dirname)).mkdir();
    	//process statistics
		
		//write to disk
		StatKeeper.saveNotableStats("./Results/"+dirname+"/stats.csv", FrevoMain.STATISTICS, true);
		StatKeeper.saveRboxFile("./Results/"+dirname+"/rboxplot.txt", FrevoMain.STATISTICS,1);
	}
	
	/** Returns the name of the component defined by the give type as human readable string */
	public static String getComponentTypeAsString(int type) {
		switch (type) {
		case FREVO_IPROBLEM: return "Problem";
		case FREVO_IMETHOD: return "Method";
		case FREVO_IREPRESENTATION: return "Representation";
		case FREVO_IRANKING: return "Ranking";
		}
		return null;
	}

	/** Saves FREVO settings to the config file */
	public static void saveSettings() {
		File configfile;
		Document doc;
		configfile = new File("frevo_config.xml");
		if (configfile.exists()) {
			doc = SafeSAX.read(configfile);
			
			Node windowsizes =  doc.selectSingleNode("/frevo/window-sizes");
			
			List<?> items = windowsizes.selectNodes(".//window");
			Iterator<?> it = items.iterator();

			while(it.hasNext()) {
				Element el = (Element) it.next();
				String name = el.valueOf("./@name");
				
				if (name.equals("main")) {
					el.addAttribute("width", Integer.toString(FrevoMain.mainWindowParameters[0]));
					el.addAttribute("height", Integer.toString(FrevoMain.mainWindowParameters[1]));
				}
				if (name.equals("componentbrowser")) {
					el.addAttribute("width", Integer.toString(FrevoMain.componentBrowserParameters[0]));
					el.addAttribute("height", Integer.toString(FrevoMain.componentBrowserParameters[1]));
					el.addAttribute("topsplit", Integer.toString(FrevoMain.componentBrowserParameters[2]));
					el.addAttribute("bigsplit", Integer.toString(FrevoMain.componentBrowserParameters[3]));
				}
			}
		} else {
			try {
				configfile.createNewFile();
			} catch (IOException e) {
				e.printStackTrace();
			}
			//file does not exist
			doc = DocumentHelper.createDocument();
			doc.addDocType("frevo", null, "config.dtd");
	        Element dfrevo = doc.addElement("frevo");
	        
	        dfrevo.addElement("tags");
	        
	        Element windowsizes = dfrevo.addElement("window-sizes");
	        Element mainwindow = windowsizes.addElement("window");
	        mainwindow.addAttribute("name", "main");
	        mainwindow.addAttribute("width", Integer.toString(FrevoMain.mainWindowParameters[0]));
	        mainwindow.addAttribute("height", Integer.toString(FrevoMain.mainWindowParameters[1]));
	        
	        Element compwindow = windowsizes.addElement("window");
	        compwindow.addAttribute("name", "componentbrowser");
	        compwindow.addAttribute("width", Integer.toString(FrevoMain.componentBrowserParameters[0]));
	        compwindow.addAttribute("height", Integer.toString(FrevoMain.componentBrowserParameters[1]));
	        compwindow.addAttribute("topsplit", Integer.toString(FrevoMain.componentBrowserParameters[2]));
	        compwindow.addAttribute("bigsplit", Integer.toString(FrevoMain.componentBrowserParameters[3]));
		}
		
		//write to file
		try {
			OutputFormat format = OutputFormat.createPrettyPrint();
            format.setLineSeparator(System.getProperty("line.separator"));
            
			configfile.createNewFile();
			FileWriter out = new FileWriter(configfile);
			BufferedWriter bw = new BufferedWriter(out);
			XMLWriter wr = new XMLWriter(bw, format);
			wr.write(doc);
			wr.close();
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}
}
